//-----------------------------------------------------------------------------
// File:     ProjectileMotion.cpp 
// Class:    None
// Parent:   None
// Children: None
// Purpose:  Simulates projectile motion of a ball
//-----------------------------------------------------------------------------
// The following are standard C/C++ header files.
// If a filename is enclosed inside < >  it means the header file is in the Include directory.
// If a filename is enclosed inside " "  it means the header file is in the current directory.
#include <ctype.h>      // Character Types
#include <math.h>       // Mathematical Constants
#include <stdarg.h>     // Variable Argument Lists
#include <stdio.h>      // Standard Input/Output Functions
#include <stdlib.h>     // Utility Functions
#include <string.h>     // String Operations
#include <signal.h>     // Signals (Contol-C + Unix System Calls)
#include <setjmp.h>     // Nonlocal Goto (For Control-C)
#include <time.h>       // Time and Date information
#include <assert.h>     // Verify Program Assertion
#include <errno.h>      // Error Codes (Used in Unix system())
#include <float.h>      // Floating Point Constants
#include <limits.h>     // Implementation Constants
#include <stddef.h>     // Standard Definitions
#include <exception>    // Exception handling (e.g., try, catch throw)
//-----------------------------------------------------------------------------
#include "SimTKsimbody.h"
using namespace SimTK;
using namespace std;
//-----------------------------------------------------------------------------
#define PI 3.1415926535897932

//-----------------------------------------------------------------------------
// Prototypes for local functions (functions not called by code in other files)
//-----------------------------------------------------------------------------
bool  SimulateProjectileMotion( void );
bool  WriteStringToFile(   const char outputString[], FILE *fptr )  { return fputs( outputString, fptr ) != 0; }
bool  WriteStringToScreen( const char outputString[] )              { return WriteStringToFile( outputString, stdout ); }
bool  WriteDoubleToFile( double x, int precision, FILE *fptr );
FILE*  FileOpenWithMessageIfCannotOpen( const char *filename, const char *attribute );


//-----------------------------------------------------------------------------
// The executable program starts here
//-----------------------------------------------------------------------------
int  main( int numberOfCommandLineArguments, char *arrayOfCommandLineArguments[] )
{
   // Simulate the multibody system
   bool simulationSucceeded = SimulateProjectileMotion();

   // Keep the screen displayed until the user presses the Enter key
   WriteStringToScreen( "\n\n Press  Enter  to terminate the program: " );
   getchar();

   // The value returned by the main function is the exit status of the program.
   // A normal program exit returns 0 (other return values usually signal an error).
   return simulationSucceeded == true ? 0 : 1;
}



//-----------------------------------------------------------------------------
bool  SimulateProjectileMotion( void )
{
   // Declare a multibody system (contains one or more force and matter sub-systems)
   MultibodySystem  mbs;

   // 0. The ground's right-handed, orthogonal x,y,z unit vectors are directed with x horizontally right and y vertically upward.
   // 1. Create a gravity vector that is straight down (in the ground's frame)
   // 2. Create a uniform gravity sub-system
   // 3. Add the gravity sub-system to the multibody system
   Vec3 gravityVector( 0, -9.8, 0 );
   UniformGravitySubsystem gravity( gravityVector );
   mbs.addForceSubsystem( gravity );

   // Create a matter sub-system (the ball)
   SimbodyMatterSubsystem  ball;

   // Create the mass, center of mass, and inertia properties for the ball
   const Real  massOfBall = 0.142;

   // The location of the ball's center of mass is a vector from the ball's
   // origin expressed in the "x, y, z"' unit vectors fixed in the ball's frame.
   // Example: The vector(0,0,0) locates the ball's center of mass at the ball's origin.
   // Example: The vector(1,0,0) locates the ball's center of mass 1 unit in the "x" direction from the ball's origin.
   const Vec3  ballCenterOfMassLocation( 0, 0, 0 );

   // Create the ball's inertia matrix about its origin for the "x, y, z" unit vectors fixed in the ball's frame.
   // Note: The 3x3 inertia matrix is symmetric - so only 6 elements need to be defined.
   // Ixx, Iyy, Izz  are moments  of inertia (    diagonal terms in the matrix)
   // Ixy, Ixz, Iyz  are products of inertia (off-diagonal terms in the matrix)
   const Real  Ixx = 1,  Iyy = 1,   Izz = 1;
   const Real  Ixy = 0,  Ixz = 0,   Iyz = 0;
   const Inertia  ballInertiaMatrix( Ixx, Iyy, Izz, Ixy, Ixz, Iyz );

   // The MassProperties class holds the mass, center of mass, and inertia properties of a rigid body.
   // Although the next line creates an instance of the MassProperties class for the ball,
   // it does not get associated with the ball until the  addRigidBody  method.
   MassProperties  ballMassProperties( massOfBall, ballCenterOfMassLocation, ballInertiaMatrix );

   // The motion of the ball is related to the motion of the ground via "mobilizers"
   // The "mobilizers" specify the allowable motion of the ball to the ground.
   // More specifically, the motion of an "outboard body" (e.g., the ball)
   // to its inboard body (e.g., the ground) is specified by first constructing:
   // 1. An  "outboard frame"  on the ground (which hooks to the "outboard body" - the ball )
   // 2. An  "inboard  frame"  on the ball  (which hooks to the "inboard  body" - the ground)

   // The orientation and position of the  outboard frame  from the ground's frame is specified below.
   // The outboard frame's axes are aligned with the ground's axes and its origin is coincident with the ground's origin.
   // In other words, for this simple problem the outboard frame and the ground frame are identical.
   const Transform outboardFrameTransformFromGround;  // The default constructor is the identity transform

   // The orientation and position of the  inboard  frame  from the ball's  frame is specified below.
   // The inboard frame's  axes are aligned with the ball's  axes and its origin is coincident with the ball's  origin
   // In other words, for this simple problem the inboard  frame and the ball frame are identical.
   // Although the inboard frame can be constructed in a simple manner analogous to the outboardFrameTransformFromGround (above)
   // it is worthwhile to look at the details of the rotation matrix and position vector in the transform:
   // a. The rotation matrix relating the InboardFrame's x,y,z axes to the BallFrame's x,y,z axes is specified InboardFrame_BallFrame.
   // b. The position of the InboardFrame's origin from the BallFrame origin, expressed in terms of the BallFrame's "x, y, z" unit vectors.
   const Rotation   inboardFrameOrientationInBall;   // ( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3       inboardFrameOriginLocationFromBallOrigin( 0, 0, 0 );
   const Transform  inboardFrameTransformFromBall( inboardFrameOrientationInBall, inboardFrameOriginLocationFromBallOrigin );

   // There are many ways that the ball can move relative to the ground.
   // The following allows the ball to move in "x", "y", and "z" directions.
   // Another option producing the same result is Mobilizer::Free
   Mobilizer ballToGroundMobilizer = Mobilizer::Cartesian;
   const BodyId ballBodyId = ball.addRigidBody( ballMassProperties, inboardFrameTransformFromBall, GroundId, outboardFrameTransformFromGround, ballToGroundMobilizer );

   // Add the matter (ball) sub-system to the system.
   mbs.setMatterSubsystem( ball );

   // Create a state for this system.
   // Define appropriate states for this multi-body system.
   // Set the initial time to 0.0
   State s;
   mbs.realize( s );
   s.setTime( 0.0 );

   // Set the initial values for the configuration variables (x,y,z)
   ball.setMobilizerQ( s, ballBodyId, 0,  0.0 );
   ball.setMobilizerQ( s, ballBodyId, 1,  0.0 );
   ball.setMobilizerQ( s, ballBodyId, 2,  0.0 );

   // Set the initial values for the motion variables
   ball.setMobilizerU( s, ballBodyId, 0,  44.7*cos(35*PI/180.0) );
   ball.setMobilizerU( s, ballBodyId, 1,  44.7*sin(35*PI/180.0) );
   ball.setMobilizerU( s, ballBodyId, 2,  0.0 );

   // Create a study using the Runge Kutta Merson integrator (alternately use the CPodesIntegrator)
   RungeKuttaMerson myStudy( mbs, s );

   // Set the numerical accuracy for the integrator
   myStudy.setAccuracy( 1.0E-7 );

   // The next statement does lots of accounting
   myStudy.initialize();

   // Open a file to record the simulation results (they are also displayed on screen)
   FILE *outputFile = FileOpenWithMessageIfCannotOpen( "ProjectileMotionResults.txt", "w" );
   WriteStringToFile(   "time      xLocation      yLocation    yVelocity  mechanicalEnergy  yDifferenceFromExact\n",  outputFile );
   WriteStringToScreen( "time      xLocation      yLocation    yVelocity  mechanicalEnergy  yDifferenceFromExact\n" );

   // Visualize results with VTKReporter
   VTKReporter animationResults( mbs );

   // For visualization purposes only, create a red sphere with a radius of 0.5 meters (huge but visible)
   DecorativeSphere redSphere = DecorativeSphere(0.5);
   redSphere.setColor(Red);     // Can also specify a Vec3 with rgb which scale from 0 to 1
   redSphere.setOpacity(0.0);   // 0.0 is solid and 1.0 is transparent

   // Decorate the ball with the red sphere at the ball's origin
   const Rotation  ballToRedSphereOrientation; //( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3      ballOriginToRedSphereOriginLocation( 0, 0, 0 );
   const Transform ballToRedSphereTransform( ballToRedSphereOrientation, ballOriginToRedSphereOriginLocation );
   animationResults.addDecoration( ballBodyId, ballToRedSphereTransform, redSphere );

   // Set the numerical integration step and the time for the simulation to run
   const Real dt = 0.01;
   const Real finalTime = 5.23;

   // Run the simulation and print the results
   while( 1 )
   {
      // Query for results to be printed
      Real time = s.getTime();
      Real kineticEnergy = mbs.getKineticEnergy(s);
      Real uniformGravitationalPotentialEnergy = mbs.getPotentialEnergy(s);
      Real mechanicalEnergy = kineticEnergy + uniformGravitationalPotentialEnergy;

      // Locate the ball origin's from the ground's origin, expressed in terms of the ground's "x,y,z" unit vectors.
      // Extract the ball's y-location from this vector.
      const Vec3 ballLocation = ball.calcBodyOriginLocationInBody( s, ballBodyId, GroundId );
      Real xLocation = ballLocation[0];
      Real yLocation = ballLocation[1];
      

      // Get the ball origin's velocity in ground, expressed in terms of the ground's "x,y,z" unit vectors.
      // Extract the ball's y-velocity from this vector.
      const Vec3 ballVelocity = ball.calcBodyOriginVelocityInBody( s, ballBodyId, GroundId );
      Real yVelocity = ballVelocity[1];

      // Exact analytical results and difference of numerical integration results
      double yExact = 10 - 4.9*time*time;
      double yDifferenceFromExact = yExact - yLocation;

      // Print results to screen
      WriteDoubleToFile( time,                 2, stdout );
      WriteDoubleToFile( xLocation,            4, stdout );
      WriteDoubleToFile( yLocation,            4, stdout );
      WriteDoubleToFile( yVelocity,            4, stdout );
      WriteDoubleToFile( mechanicalEnergy,     7, stdout );
      WriteDoubleToFile( yDifferenceFromExact, 7, stdout );
      WriteStringToScreen( "\n" );
      // Print results to file
      WriteDoubleToFile( time,                 2, outputFile );
      WriteDoubleToFile( xLocation,            4, outputFile );
      WriteDoubleToFile( yLocation,            4, outputFile );
      WriteDoubleToFile( yVelocity,            4, outputFile );
      WriteDoubleToFile( mechanicalEnergy,     7, outputFile );
      WriteDoubleToFile( yDifferenceFromExact, 7, outputFile );
      WriteStringToFile( "\n", outputFile );

      // Animate the results for this step
      animationResults.report(s);

      // Check if integration has completed
      if( time >= finalTime ) break;

      // Increment time step
      myStudy.step( time + dt);
   }

   // Simulation completed properly
   return true;
}


//-----------------------------------------------------------------------------
FILE*  FileOpenWithMessageIfCannotOpen( const char *filename, const char *attribute )
{
   // Try to open the file
   FILE *Fptr1 = fopen( filename, attribute );

   // If unable to open the file, issue a message
   if( !Fptr1 )
   {
      WriteStringToScreen( "\n\n Unable to open the file: " );
      WriteStringToScreen( filename );
      WriteStringToScreen( "\n\n" );
   }

   return Fptr1;
}


//-----------------------------------------------------------------------------
bool  WriteDoubleToFile( double x, int precision, FILE *fptr )
{
   // Ensure the precision (number of digits in the mantissa after the decimal point) makes sense.
   // Next, calculate the field width so it includes one extra space to the right of the number.
   if( precision < 0 || precision > 17 ) precision = 5;
   int fieldWidth = precision + 8;

   // Create the format specifier and print the number
   char format[20];
   sprintf( format, " %%- %d.%dE", fieldWidth, precision );
   return fprintf( fptr, format, x ) >= 0;
}


