//-----------------------------------------------------------------------------
// File:     RotatingBook.cpp 
// Class:    None
// Parent:   None
// Children: None
// Purpose:  Simulates rotating motion of a book
//-----------------------------------------------------------------------------
// The following are standard C/C++ header files.
// If a filename is enclosed inside < >  it means the header file is in the Include directory.
// If a filename is enclosed inside " "  it means the header file is in the current directory.
#include <ctype.h>      // Character Types
#include <math.h>       // Mathematical Constants
#include <stdarg.h>     // Variable Argument Lists
#include <stdio.h>      // Standard Input/Output Functions
#include <stdlib.h>     // Utility Functions
#include <string.h>     // String Operations
#include <signal.h>     // Signals (Contol-C + Unix System Calls)
#include <setjmp.h>     // Nonlocal Goto (For Control-C)
#include <time.h>       // Time and Date information
#include <assert.h>     // Verify Program Assertion
#include <errno.h>      // Error Codes (Used in Unix system())
#include <float.h>      // Floating Point Constants
#include <limits.h>     // Implementation Constants
#include <stddef.h>     // Standard Definitions
#include <exception>    // Exception handling (e.g., try, catch throw)
//-----------------------------------------------------------------------------
#include "SimTKsimbody.h"
using namespace SimTK;
using namespace std;
//-----------------------------------------------------------------------------
#define PI 3.141592653589793238462643

//-----------------------------------------------------------------------------
// Prototypes for local functions (functions not called by code in other files)
//-----------------------------------------------------------------------------
bool  SimulateRotatingBook( void );
bool  WriteStringToFile(   const char outputString[], FILE *fptr )  { return fputs( outputString, fptr ) != 0; }
bool  WriteStringToScreen( const char outputString[] )              { return WriteStringToFile( outputString, stdout ); }
bool  WriteDoubleToFile( double x, int precision, FILE *fptr );
FILE*  FileOpenWithMessageIfCannotOpen( const char *filename, const char *attribute );


//-----------------------------------------------------------------------------
// The executable program starts here
//-----------------------------------------------------------------------------
int  main( int numberOfCommandLineArguments, char *arrayOfCommandLineArguments[] )
{
   // Simulate the multibody system
   bool simulationSucceeded = SimulateRotatingBook();

   // Keep the screen displayed until the user presses the Enter key
   WriteStringToScreen( "\n\n Press  Enter  to terminate the program: " );
   getchar();

   // The value returned by the main function is the exit status of the program.
   // A normal program exit returns 0 (other return values usually signal an error).
   return simulationSucceeded == true ? 0 : 1;
}



//-----------------------------------------------------------------------------
bool  SimulateRotatingBook( void )
{
   // Declare a multibody system (contains one or more force and matter sub-systems)
   MultibodySystem  mbs;

   // 0. The ground's right-handed, orthogonal x,y,z unit vectors are directed with x horizontally right and y vertically upward.
   // 1. Create a gravity vector that is straight down (in the ground's frame)
   // 2. Create a uniform gravity sub-system
   // 3. Add the gravity sub-system to the multibody system
   Vec3 gravityVector( 0, -9.8, 0 );
   UniformGravitySubsystem gravity( gravityVector );
   mbs.addForceSubsystem( gravity );

   // Create a matter sub-system (the book)
   SimbodyMatterSubsystem  book;

   // Create the mass, center of mass, and inertia properties for the book
   const Real  massOfBook = 0.4;

   // The location of the book's center of mass is a vector from the book's
   // origin expressed in the "x, y, z"' unit vectors fixed in the book's frame.
   // Example: The vector(0,0,0) locates the book's center of mass at the book's origin.
   // Example: The vector(1,0,0) locates the book's center of mass 1 unit in the "x" direction from the book's origin.
   const Vec3  bookCenterOfMassLocation( 0, 0, 0 );

   // Create the book's inertia matrix about its origin for the "x, y, z" unit vectors fixed in the book's frame.
   // Note: The 3x3 inertia matrix is symmetric - so only 6 elements need to be defined.
   // Ixx, Iyy, Izz  are moments  of inertia (    diagonal terms in the matrix)
   // Ixy, Ixz, Iyz  are products of inertia (off-diagonal terms in the matrix)
   const Real a = 0.05, b = 0.30, c = 0.20;
   const Real  Ixx = 1/12.0*massOfBook*(a*a+b*b),  Iyy = 1/12.0*massOfBook*(a*a+c*c),   Izz = 1/12.0*massOfBook*(c*c+b*b);
   const Real  Ixy = 0,  Ixz = 0,   Iyz = 0;
   const Inertia  bookInertiaMatrix( Ixx, Iyy, Izz, Ixy, Ixz, Iyz );

   // The MassProperties class holds the mass, center of mass, and inertia properties of a rigid body.
   // Although the next line creates an instance of the MassProperties class for the book,
   // it does not get associated with the book until the  addRigidBody  method.
   MassProperties  bookMassProperties( massOfBook, bookCenterOfMassLocation, bookInertiaMatrix );

   // The motion of the book is related to the motion of the ground via "mobilizers"
   // The "mobilizers" specify the allowable motion of the book to the ground.
   // More specifically, the motion of an "outboard body" (e.g., the book)
   // to its inboard body (e.g., the ground) is specified by first constructing:
   // 1. An  "outboard frame"  on the ground (which hooks to the "outboard body" - the book )
   // 2. An  "inboard  frame"  on the book  (which hooks to the "inboard  body" - the ground)

   // The orientation and position of the  outboard frame  from the ground's frame is specified below.
   // The outboard frame's axes are aligned with the ground's axes and its origin is coincident with the ground's origin.
   // In other words, for this simple problem the outboard frame and the ground frame are identical.
   const Transform outboardFrameTransformFromGround;  // The default constructor is the identity transform

   // The orientation and position of the  inboard  frame  from the book's  frame is specified below.
   // The inboard frame's  axes are aligned with the book's  axes and its origin is coincident with the book's  origin
   // In other words, for this simple problem the inboard  frame and the book frame are identical.
   // Although the inboard frame can be constructed in a simple manner analogous to the outboardFrameTransformFromGround (above)
   // it is worthwhile to look at the details of the rotation matrix and position vector in the transform:
   // a. The rotation matrix relating the InboardFrame's x,y,z axes to the bookFrame's x,y,z axes is specified InboardFrame_BookFrame.
   // b. The position of the InboardFrame's origin from the BookFrame origin, expressed in terms of the BookFrame's "x, y, z" unit vectors.
   const Rotation   inboardFrameOrientationInBook;   // ( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3       inboardFrameOriginLocationFromBookOrigin( 0, 0, 0 );
   const Transform  inboardFrameTransformFromBook( inboardFrameOrientationInBook, inboardFrameOriginLocationFromBookOrigin );

   // There are many ways that the book can move relative to the ground.
   // The following allows the book to move in "x", "y", and "z" directions.
   // Another option producing the same result is Mobilizer::Free
   Mobilizer bookToGroundMobilizer = Mobilizer::Ball;
   const BodyId bookBodyId = book.addRigidBody( bookMassProperties, inboardFrameTransformFromBook, GroundId, outboardFrameTransformFromGround, bookToGroundMobilizer );

   // Add the matter (book) sub-system to the system.
   mbs.setMatterSubsystem( book );

   // Create a state for this system.
   // Define appropriate states for this multi-body system.
   // Set the initial time to 0.0
   State s;
   mbs.realize( s );
   s.setTime( 0.0 );

   // Set the initial values for the configuration variables (x,y,z)
   book.setMobilizerQ( s, bookBodyId, 0,  1.0 );
   book.setMobilizerQ( s, bookBodyId, 1,  0.0 );
   book.setMobilizerQ( s, bookBodyId, 2,  0.0 );
   book.setMobilizerQ( s, bookBodyId, 3,  0.0 );

   // Set the initial values for the motion variables
   //const Real wx = 0.2, wy = 7.0, wz = 0.2;  // Spin about minimum axis 
   //const Real wx = 7.0, wy = 0.2, wz = 0.2;  // Spin about intermediate axis 
   const Real wx = 0.2, wy = 0.2, wz = 7.0;  // Spin about maximum axis 

   book.setMobilizerU( s, bookBodyId, 0,  wx );
   book.setMobilizerU( s, bookBodyId, 1,  wy );
   book.setMobilizerU( s, bookBodyId, 2,  wz );

   // Create a study using the Runge Kutta Merson integrator (alternately use the CPodesIntegrator)
   RungeKuttaMerson myStudy( mbs, s );

   // Set the numerical accuracy for the integrator
   myStudy.setAccuracy( 1.0E-7 );

   // The next statement does lots of accounting
   myStudy.initialize();

   // Open a file to record the simulation results (they are also displayed on screen)
   FILE *outputFile = FileOpenWithMessageIfCannotOpen( "RotatingBookResults.txt", "w" );
   WriteStringToFile(   "  time          wx           wy           wz            Hx            Hy            Hz          magH          K\n",  outputFile );
   WriteStringToScreen( "  time          wx           wy           wz            Hx            Hy            Hz          magH          K\n" );

   // Visualize results with VTKReporter
   VTKReporter animationResults( mbs );

   // For visualization purposes only, create a red brick with a radius of 0.5 meters (huge but visible)
   const Vec3 halfLengths( c/2.0, b/2.0, a/2.0 );
   DecorativeBrick redBrick = DecorativeBrick( halfLengths );
   redBrick.setColor(Red);     // Can also specify a Vec3 with rgb which scale from 0 to 1
   redBrick.setOpacity(0.0);   // 0.0 is solid and 1.0 is transparent

   // Decorate the book with the red brick at the book's origin
   const Rotation  bookToRedBrickOrientation; //( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3      bookOriginToRedBrickOriginLocation( 0, 0, 0 );
   const Transform bookToRedBrickTransform( bookToRedBrickOrientation, bookOriginToRedBrickOriginLocation );
   animationResults.addDecoration( bookBodyId, bookToRedBrickTransform, redBrick );

   // Set the numerical integration step and the time for the simulation to run
   const Real dt = 0.01;
   const Real finalTime = 4.0;

   // Run the simulation and print the results
   while( 1 )
   {
      // Query for results to be printed
      Real time = s.getTime();
      Real kineticEnergy = mbs.getKineticEnergy(s);
      Real uniformGravitationalPotentialEnergy = mbs.getPotentialEnergy(s);
      Real mechanicalEnergy = kineticEnergy + uniformGravitationalPotentialEnergy;

      // Locate the book origin's from the ground's origin, expressed in terms of the ground's "x,y,z" unit vectors.
      // Extract the book's y-location from this vector.
      const Vec3 bookLocation = book.calcBodyOriginLocationInBody( s, bookBodyId, GroundId );
      Real xLoc = bookLocation[0];
      Real yLoc = bookLocation[1];
      Real zLoc = bookLocation[2];

      // Get the book origin's velocity in ground, expressed in terms of the ground's "x,y,z" unit vectors.
      // Extract the book's y-velocity from this vector.

      const Vec3 bookVelocity = book.calcBodyOriginVelocityInBody( s, bookBodyId, GroundId );
      Real xLinVelocity = bookVelocity[0];
      Real yLinVelocity = bookVelocity[1];
      Real zLinVelocity = bookVelocity[2];

      const Vec3 bookAngularVelocityExpressedInGround = book.calcBodyAngularVelocityInBody( s, bookBodyId, GroundId );
      Real xAngVelocity = bookAngularVelocityExpressedInGround[0];
      Real yAngVelocity = bookAngularVelocityExpressedInGround[1];
      Real zAngVelocity = bookAngularVelocityExpressedInGround[2];

      // Get rotation matrix from ground to book & invert
      Rotation rotationFromGroundToBook = book.getBodyRotation( s, bookBodyId );
      Rotation rotationFromBookToGround = rotationFromGroundToBook.invert();

      Vec3 angularVelocityExpressedInBook = rotationFromBookToGround*bookAngularVelocityExpressedInGround;

      Real angVelX = angularVelocityExpressedInBook[0];
      Real angVelY = angularVelocityExpressedInBook[1];
      Real angVelZ = angularVelocityExpressedInBook[2];

      // Find angular momentum of book about its center of mass.
      const SpatialVec H_CB = book.calcBodyMomentumAboutBodyMassCenterInGround( s, bookBodyId);
	   
      Real H_CB_x = Ixx*angVelX;
      Real H_CB_y = Iyy*angVelY;
      Real H_CB_z = Izz*angVelZ;

      Real magH = sqrt( H_CB_x*H_CB_x + H_CB_y*H_CB_y + H_CB_z*H_CB_z );

      // Print results to screen
      WriteDoubleToFile( time,                 2, stdout );
      WriteDoubleToFile( angVelX,              4, stdout );
      WriteDoubleToFile( angVelY,              4, stdout );
      WriteDoubleToFile( angVelZ,              4, stdout );
      WriteDoubleToFile( H_CB_x,               4, stdout );
      WriteDoubleToFile( H_CB_y,               4, stdout );
      WriteDoubleToFile( H_CB_z,               4, stdout );
      WriteDoubleToFile( magH,                 4, stdout );
      WriteDoubleToFile( kineticEnergy,        7, stdout );
      WriteStringToScreen( "\n" );

      // Print results to file
      WriteDoubleToFile( time,                 2, outputFile );
      WriteDoubleToFile( angVelX,              4, outputFile );
      WriteDoubleToFile( angVelY,              4, outputFile );
      WriteDoubleToFile( angVelZ,              4, outputFile );
      WriteDoubleToFile( H_CB_x,               4, outputFile );
      WriteDoubleToFile( H_CB_y,               4, outputFile );
      WriteDoubleToFile( H_CB_z,               4, outputFile );
      WriteDoubleToFile( magH,                 4, outputFile );
      WriteDoubleToFile( kineticEnergy,        7, outputFile );
      WriteStringToFile( "\n", outputFile );

      // Animate the results for this step
      animationResults.report(s);

      // Check if integration has completed
      if( time >= finalTime ) break;

      // Increment time step
      myStudy.step( time + dt);
   }

   // Simulation completed properly
   return true;
}


//-----------------------------------------------------------------------------
FILE*  FileOpenWithMessageIfCannotOpen( const char *filename, const char *attribute )
{
   // Try to open the file
   FILE *Fptr1 = fopen( filename, attribute );

   // If unable to open the file, issue a message
   if( !Fptr1 )
   {
      WriteStringToScreen( "\n\n Unable to open the file: " );
      WriteStringToScreen( filename );
      WriteStringToScreen( "\n\n" );
   }

   return Fptr1;
}


//-----------------------------------------------------------------------------
bool  WriteDoubleToFile( double x, int precision, FILE *fptr )
{
   // Ensure the precision (number of digits in the mantissa after the decimal point) makes sense.
   // Next, calculate the field width so it includes one extra space to the right of the number.
   if( precision < 0 || precision > 17 ) precision = 5;
   int fieldWidth = precision + 8;

   // Create the format specifier and print the number
   char format[20];
   sprintf( format, " %%- %d.%dE", fieldWidth, precision );
   return fprintf( fptr, format, x ) >= 0;
}


